##################################################################
# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004 ObjectRealms, LLC
# Copyright @ 2004 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

"""
generic xml marshaller

 based on registering namespaces with the marshaller,
 the marshaller tries to delegate as much as possible
 to the namespaces, the default implementation of which
 delegates as much as possible to schema attributes
 within that namespace.

 see the Marshall.namespaces package for some sample and
 default namespaces.

caveats

 - if you want to use multiple namespaces on the same
   xml node, then this isn't the parser for you. you
   can do some basic hacks around it w/ ParseContext
   namespace delegation.

Authors: kapil thangavelu <k_vertigo@objectrealms.net> (current impl)
         sidnei da silva <sidnei@awkly.org>

"""

#################################
import sys
import thread
import traceback
from xml.dom import minidom
from xml.etree import cElementTree as ElementTree
from Products.Archetypes.Marshall import Marshaller
from Products.Marshall import config
from Products.Marshall import utils

#################################

_marker = object()

XMLNS_NS = 'http://www.w3.org/2000/xmlns/'
XMLREADER_START_ELEMENT_NODE_TYPE = 1
XMLREADER_END_ELEMENT_NODE_TYPE = 15
XMLREADER_TEXT_ELEMENT_NODE_TYPE = 3


class SchemaAttributeDemarshallException(Exception):
    """Exception mus be raised when demershall of SchemaAtribute fails."""


class ErrorCallback:

    def __init__(self):
        self.msgs = {}

    def __call__(self, ctx, msg):
        self.append(msg)

    def append(self, msg):
        tid = thread.get_ident()
        msgs = self.msgs.setdefault(tid, [])
        msgs.append(msg)

    def get(self, clear=False):
        tid = thread.get_ident()
        msgs = self.msgs.setdefault(tid, [])
        if clear:
            self.clear()
        return ''.join(msgs)

    def clear(self):
        tid = thread.get_ident()
        self.msgs[tid] = []


class XmlNamespace(object):

    #################################
    # the framework does a bit of introspection on
    # namespaces for some attributes, defined below

    # whether or not this namespace uses fields from an
    # object's at schema. if true then this namespace
    # should also define the get getATFields below
    uses_at_fields = False

    # the xml namespace uri
    xmlns = "http://example.com"

    # the xml namespace prefix
    prefix = "xxx"
    #################################

    def __init__(self):
        for attribute in self.attributes:
            attribute.setNamespace(self)

    def getAttributeByName(self, name):
        """ given an xml name return the schema attribute
        """
        for attribute in self.attributes:
            if attribute.name == name:
                return attribute
        return None

    def getRelaxNG(self):
        """ get the relaxng fragment that defines
        whats in the namespace
        """
        raise NotImplemented("Subclass Responsiblity")

    def getATFields(self):
        """ return the at schema field names which are
        handled by this namespace, this is utilized by
        the AT namespace so it doesn't also handle these
        fields. """
        raise NotImplemented("Subclass Responsiblity")

    def serialize(self, dom_node, parent_node, instance, options):
        """ serialize the instance values to xml
        based on attributes in this namespace
        """
        for attribute in self.attributes:
            attribute.serialize(dom_node, parent_node, instance)

    def deserialize(self, instance, ns_data, options):
        """ given the instance and the namespace data for
        instance, reconstitute this namespace's attributes
        on the instance.
        """
        if not ns_data:
            return
        for attribute in self.attributes:
            try:
                attribute.deserialize(instance, ns_data)
            except Exception, e:
                ec, e, tb = sys.exc_info()
                ftb = traceback.format_tb(tb,)
                msg = "failure while demarshalling schema attribute %s\n" % \
                      attribute.name
                msg += "data: %s\n" % ns_data.get(attribute.name, None)
                msg += "original exception: %s\n" % str(ec)
                msg += "original traceback:\n%s" % '\n'.join(ftb)
                raise SchemaAttributeDemarshallException(msg)

    def processXml(self, context, node):
        """ handle the start of a xml tag with this namespace
        the namespace and the name of the tag are bound to node.

        if this method return false then the node is assumed to
        be junk and it is discarded.
        """
        tag, ns = utils.fixtag(node.tag, context.ns_map)
        attribute = self.getAttributeByName(tag)
        if attribute is None:
            return False
        node.set('attribute', attribute)
        return attribute.processXml(context, node)

    def processXmlEnd(self, name, context):
        """ callback invoked when the parser reaches the
        end of an xml node in this namespace.
        """

    def getSchemaInfo(self):
        """ return information on this namespace's rng schema

        should be an iterable of sets of
        ('defined_name', 'occurence', 'schema')
        where defined name is the name of any top level defined
        entities in the schema, occurence defines the rng occurence
        value for that entity in the object's xml representation,
        and schema is the rng schema definition for the defined entities
        """
        return ()


class SchemaAttribute(object):

    def __init__(self, name, field_name=None):
        self.name, self.field_id = name, field_name or name
        self.namespace = None

    def set(self, instance, data):
        """ set the attribute's value on the instance
        """
        raise NotImplemented

    def get(self, instance):
        """ retrieve the schema attribute's value from the instance
        """
        raise NotImplemented

    def serialize(self, dom, instance):
        """ serialize the attribute's instance value into the dom
        """
        raise NotImplemented

    def deserialize(self, instance, ns_data):
        """ give the instance and the namespace data for
        instance, reconstitute this attribute on the instance
        """
        self.set(instance, ns_data)

    def processXml(self, context, ctx_node):
        """ callback invoked with a node from the xml stream
        if false is returned the current node is assumed to be
        not interesting and is ignored.
        """
        return True

    def processXmlValue(self, context, value):
        """ callback to process text nodes
        """
        value = value and value.strip()
        if not value:
            return
        data = context.getDataFor(self.namespace.xmlns)
        data[self.name] = value

    def setNamespace(self, namespace):
        """ sets which namespace the attribute belongs to
        """
        self.namespace = namespace


class DataNode(object):
    """ a data bag holding a namespace uri and a node name
    """
    __slots__ = (
        'ns',
        'name',
        'attribute',
        )

    def __init__(self, ns, name):
        self.ns = ns
        self.name = name
        self.attribute = None


class ParseContext(object):
    """ a bag for holding data values from and for parsing
    """
    def __init__(self, instance, root, ns_map):
        self.instance = instance
        self.root = root # root node
        self.ns_map = ns_map # ns_uri -> namepace
        self.data = {} # ns_uri -> ns_data
        self.node = None # current node if any
        self.ns_delegate = None

    def getDataFor(self, ns_uri):
        return self.data.setdefault(ns_uri, {})

    def getNamespaceFor(self, ns_uri):
        if self.ns_delegate is not None:
            return self.ns_delegate
        return self.ns_map.get(ns_uri)

    def setNamespaceDelegate(self, namespace):
        self.ns_delegate = namespace


class ATXMLMarker:
    pass

_marker = ATXMLMarker()


class ATXMLMarshaller(Marshaller):

    # Just a plain list of ns objects.
    namespaces = []

    # options for a subclass
    use_validation = False

    __name__ = 'ATXML Marshaller'

    def __init__(self, demarshall_hook=None, marshall_hook=None,
                 root_tag='metadata', namespace=_marker):
        Marshaller.__init__(self, demarshall_hook, marshall_hook)
        self.root_tag = root_tag
        self.namespace = namespace
        if namespace is _marker:
            self.namespace = config.AT_NS

    def getFieldNamespace(self, field):
        # Flatten ns into (ns, attr) tuples
        flat_ns = []
        [flat_ns.extend(zip((n,) * len(n.attrs), n.attrs)) for
         n in self.namespaces]
        # Dict mapping an AT fieldname to a (prefix, element name) tuple
        field_map = dict([(a.field, (n.prefix, a.name)) for n, a in flat_ns])
        return field_map

    def getNamespaceURIMap(self):
        """ Mapping of xmlns URI to ns object
        """
        ns_map = dict([(ns.xmlns, ns) for ns in self.namespaces])
        return ns_map

    def getNamespacePrefixMap(self):
        """ Mapping of prefix -> xmlns URI
        """
        return dict([(ns.prefix, ns.xmlns) for ns in self.namespaces])

    def getNamespaces(self, namespaces=None):
        if namespaces is None:
            for ns in getRegisteredNamespaces():
                yield ns
            raise StopIteration

        ns = getRegisteredNamespaces()
        for n in ns:
            if n.prefix in namespaces or \
               n.xmlns in namespaces:
                yield n

    def demarshall(self, instance, data, **kwargs):
        context = self.parseContext(instance, data)
        self.processContext(instance, context, kwargs)

    def marshall(self, instance, use_namespaces=None, **kwargs):
        doc = minidom.Document()
        node = doc.createElementNS(self.namespace, self.root_tag)
        doc.appendChild(node)

        # setup default namespace
        attr = doc.createAttribute('xmlns')
        attr.value = self.namespace
        node.setAttributeNode(attr)

        for ns in self.getNamespaces(use_namespaces):
            ns.serialize(doc, node, instance, kwargs)
            if not ns.prefix:
                continue
            attrname = 'xmlns:%s' % ns.prefix
            attr = doc.createAttribute(attrname)
            attr.value = ns.xmlns
            node.setAttributeNode(attr)

        content_type = 'text/xml'
        data = doc.toprettyxml()#.encode('utf-8')
        length = len(data)
        return (content_type, length, data)

    def parseContext(self, instance, data):
        #parser = XmlParser(instance, data, use_validation=self.use_validation)
        root = ElementTree.fromstring(data)
        ns_map = self.getNamespaceURIMap()
        context = ParseContext(instance, root, ns_map)
        context.xmlsource = data
        self.parseXml(root, context, ns_map)
        return context

    def parseXml(self, root, context, ns_map):
        """
        input read and dispatch loop
        """
        read_result = 1
        for node in root:
            tag, namespace = utils.fixtag(node.tag, context.ns_map)
            if namespace.processXml(context, node):
                context.node = node
                context.node.get('attribute').processXmlValue(context,
                                                              node.text)
            else:
                ## XXX: raise a warning that the attribute isnt defined
                ## XXX: in the schema
                pass

        return read_result

    def processContext(self, instance, context, options):
        """ instantiate instance with data from context
        """
        for ns in getRegisteredNamespaces():
            ns_data = context.getDataFor(ns.xmlns)
            ns.deserialize(instance, ns_data, options)


class _NamespaceCatalog(object):

    def __init__(self):
        self._namespaces = {}
        self._order = []

    def registerNamespace(self, namespace, override=False, position=-1):
        if namespace.xmlns in self._namespaces and not override:
            raise RuntimeError("Duplicate Namespace Registration %s" %
                               namespace.xmlns)
        self._namespaces[namespace.xmlns] = namespace
        if position == -1:
            position = len(self._order)
        self._order.append(position, namespace.xmlns)

    def getRegisteredNamespaces(self):
        return [self._namespaces[xmlns] for xmlns in self._order]


NamespaceCatalog = _NamespaceCatalog()

registerNamespace = NamespaceCatalog.registerNamespace
getRegisteredNamespaces = NamespaceCatalog.getRegisteredNamespaces


def registerNamespace(namespace):
    if not isinstance(namespace, XmlNamespace):
        namespace = namespace()
    ATXMLMarshaller.namespaces.append(namespace)


def getRegisteredNamespaces():
    return tuple(ATXMLMarshaller.namespaces)



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id$
"""
import os
import logging
from App.Common import package_home

logger = logging.getLogger('Marshall')
hasLibxml2 = False
hasElementtree = True

PACKAGE_HOME = package_home(globals())
HANDLE_REFS = False

TOOL_ID = 'marshaller_registry'
AT_NS = 'http://plone.org/ns/archetypes/'
CMF_NS = 'http://cmf.zope.org/namespaces/default/'
DC_NS = "http://purl.org/dc/elements/1.1/"
ADOBE_NS = "adobe:ns:meta"

ATXML_SCHEMA = os.path.join(PACKAGE_HOME, 'validation', 'atxml')



##################################################################
# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004 ObjectRealms, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

"""
Serialize CMF Framework Attributes - Type, Workflow,
Local Roles.

Created: 10/11/2004
Author: Kapil Thangavelu <k_vertigo@objectrealms.net>
"""

from DateTime import DateTime
from Products.Marshall import config
from Products.Marshall.handlers.atxml import XmlNamespace
from Products.Marshall.handlers.atxml import SchemaAttribute
from Products.Marshall import utils

TypeRNGSchemaFragment = '''
  <define name="TypeInfo"
          xmlns:cmf="http://cmf.zope.org/namespaces/default"
          datatypeLibrary="http://www.w3.org/2001/XMLSchema-datatypes"
          xmlns="http://relaxng.org/ns/structure/1.0">
           <element name="cmf:type">
             <text />
           </element>
          </zeroOrMore>
  </define>
'''


class TypeAttribute(SchemaAttribute):

    def getAttributeNames(self):
        return (self.name,)

    def get(self, instance):
        return instance.getPortalTypeName()

    def deserialize(self, instance, ns_data):
        value = ns_data.get(self.name)
        instance._setPortalTypeName(value)

    def serialize(self, dom, parent_node, instance):
        value = self.get(instance)
        elname = "%s:%s" % (self.namespace.prefix, self.name)
        node = dom.createElementNS(CMF.xmlns, elname)
        value_node = dom.createTextNode(str(value))
        node.appendChild(value_node)
        node.normalize()
        parent_node.appendChild(node)


SecurityRNGSchemaFragment = '''
  <define name="SecurityInfo"
          xmlns:zs="http://xml.zope.org/ns/local_roles"
          datatypeLibrary="http://www.w3.org/2001/XMLSchema-datatypes"
          xmlns="http://relaxng.org/ns/structure/1.0">
          <zeroOrMore>
            <element name="zs:security">
              <zeroOrMore>
               <element name="zs:local_role">
                 <attribute name="user_id" />
                 <attribute name="role" />
               </element>
              </zeroOrMore>
            </element>
          </zeroOrMore>
  </define>
'''


class LocalRolesAttribute(SchemaAttribute):
    """
    serializes local roles into the form of

    <security>

      <local_role user_id="matz" role"bar" />
      <local_role user_id="john" role"manager" />
      <local_role user_id="john" role"owner" />
      <local_role user_id="mike" role"manager" />

    </security>

    """

    def getAttributeNames(self):
        return (self.name,)

    def get(self, instance):
        return getattr(instance, '__ac_local_roles__', {})

    def deserialize(self, instance, ns_data):
        values = ns_data.get(self.name)
        if not values:
            return
        for user_id, role in values:
            instance.manage_addLocalRoles(user_id, [role])

    def serialize(self, dom, parent_node, instance):
        values = self.get(instance)

        if not values:
            return

        elname = "%s:%s" % (self.namespace.prefix, "security")
        node = dom.createElementNS(CMF.xmlns, elname)

        for user_id, roles in values.items():
            for role in roles:
                elname = "%s:%s" % (self.namespace.prefix, self.name)
                lr_node = dom.createElementNS(CMF.xmlns, elname)
                user_attr = dom.createAttribute("user_id")
                user_attr.value = user_id
                lr_node.setAttributeNode(user_attr)

                role_attr = dom.createAttribute("role")
                role_attr.value = role
                lr_node.setAttributeNode(role_attr)
                node.appendChild(lr_node)

        parent_node.appendChild(node)

    def processXml(self, context, ctx_node):
        value = context.reader.Value()
        if value is None:
            return
        value = value.strip()
        if not value:
            return

        data = context.getDataFor(self.namespace.xmlns)
        values = data.setdefault(self.name, [])
        user_id = role = None

        while context.reader.MoveToNextAttribute():
            if context.reader.LocalName() == 'user_id':
                user_id = context.reader.Value()
            elif context.reader.LocalName() == 'role':
                role = context.reader.Value()

        assert user_id, role
        values.append((user_id, role))


WorkflowRNGSchemaFragment = '''
  <define name="WorkflowHistory"
          xmlns:cmf="http://cmf.zope.org/namespaces/default/"
          datatypeLibrary="http://www.w3.org/2001/XMLSchema-datatypes"
          xmlns="http://relaxng.org/ns/structure/1.0">

      <element name="workflow_history">
       <oneOrMore>
        <element name="workflow">
          <attribute name="id" />
           <element name="history">
            <oneOrMore>
             <element name="wf_var">
              <attribute name="id" />
              <attribute name="type" />
              <attribute name="value" />
             </element>
            </oneOrMore>
           </element>
        </element>
       </oneOrMore>
      </element>

  </define>
'''


class WorkflowAttribute(SchemaAttribute):
    """
    serializes workflow state into the form below

    it will try to deal as best as possible with arbitrary
    python values, if it can't deal it will error out.

    <workflow_history>
      <workflow id="">
       <!-- in chronological order -->
       <history>
         <wf_var id="" type="" value="" />
         <wf_var id="" type="" value="" />
       </history>
      </workflow>
    </workflow_history>
    """

    def getAttributeNames(self):
        return ("workflow", "var", "history", 'workflow_history')

    def get(self, instance):
        return getattr(instance, 'workflow_history', None)

    def deserialize(self, instance, ns_data):
        wf_records = ns_data.get(self.name, {})
        for wf_id in wf_records:
            history = []
            for record in wf_records[wf_id]:
                history.append(record)
            instance.workflow_history[wf_id] = tuple(history)

    def serialize(self, dom, parent_node, instance):
        history = self.get(instance)

        if history is None:
            return

        elname = "%s:workflow_history" % (self.namespace.prefix)
        node = dom.createElementNS(self.namespace.xmlns, elname)
        keys = history.keys()
        for wf_id in keys:
            wf_node = self.serializeWorkflow(dom, wf_id, history)
            node.appendChild(wf_node)
        parent_node.appendChild(node)

    def serializeWorkflow(self, dom, wf_id, wf_hist):

        prefix = self.namespace.prefix
        xmlns = self.namespace.xmlns

        elname = "%s:workflow" % prefix
        node = dom.createElementNS(xmlns, elname)

        wfid_attr = dom.createAttribute("id")
        wfid_attr.value = wf_id

        node.setAttributeNode(wfid_attr)

        for history in wf_hist[wf_id]:
            elname = "%s:%s" % (prefix, "history")
            history_node = dom.createElementNS(xmlns, elname)
            items = history.items()
            items.sort(lambda a, b: cmp(a[0], b[0]))

            for k, v in items:
                elname = "%s:%s" % (prefix, "var")
                var_node = dom.createElementNS(xmlns, elname)
                # Attributes normally do not have a namespace
                value_attr = dom.createAttribute("value")
                type_attr = dom.createAttribute("type")
                id_attr = dom.createAttribute("id")

                value, vtype = marshall_value(v)

                id_attr.value = str(k)
                type_attr.value = vtype
                value_attr.value = value

                var_node.setAttributeNode(id_attr)
                var_node.setAttributeNode(type_attr)
                var_node.setAttributeNode(value_attr)

                history_node.appendChild(var_node)

            node.appendChild(history_node)

        return node

    def processXml(self, context, node):

        tag, namespace = utils.fixtag(node.tag, context.ns_map)
        data = context.getDataFor(self.namespace.xmlns)
        nsprefix = node.tag[:node.tag.find('}') + 1]

        #iworkflow
        wf_node = node.find(nsprefix + 'workflow')
        wf_id = (wf_node.attrib.get(nsprefix + 'id') or
                 wf_node.attrib.get('id'))
                 #be tolerant with namespace sloppyness;)
        assert wf_id

        wf_data = data.setdefault(self.name, {})
        wf_data.setdefault(wf_id, [])
        wf_pstate = data.setdefault('_wf_pstate', wf_id)

        #history
        hist_nodes = wf_node.findall(nsprefix + 'history')
        wf_pstate = data['_wf_pstate']
        for hist_node in hist_nodes:
            record = {}
            data[self.name][wf_pstate].append(record)

            #var
            var_nodes = hist_node.findall(nsprefix + 'var')
            vid = vtype = value = None

            for var_node in var_nodes:
                vid = (var_node.attrib.get(nsprefix + 'id') or
                       var_node.attrib.get('id'))
                vtype = (var_node.attrib.get(nsprefix + 'type', None) or
                         var_node.attrib.get('type'))
                value = (var_node.attrib.get(nsprefix + 'value', None) or
                         var_node.attrib.get('value') or '')

                assert vid and vtype and not value is None

                value = demarshall_value(value, vtype)
                wf_pstate = data['_wf_pstate']
                data[self.name][wf_pstate][-1][vid] = value

        return True


def marshall_value(value):

    if isinstance(value, str):
        return value, 'str'
    elif isinstance(value, int):
        return str(value), 'int'
    elif isinstance(value, float):
        return str(value), 'float'
    elif isinstance(value, DateTime):
        return value.ISO8601(), 'date'
    elif isinstance(value, type(None)):
        return 'None', 'None'
    else:
        raise SyntaxError("Unknown value type %r" % value)


def demarshall_value(value, type):

    if type == 'str':
        return value
    elif type == 'int':
        return int(value)
    elif type == 'float':
        return float(value)
    elif type == 'date':
        if value.strip() == '':
            return None
        else:
            return DateTime(value)
    elif type == 'None':
        return None
    else:
        raise SyntaxError("Unknown Type %r" % type)


class CMF(XmlNamespace):

    xmlns = config.CMF_NS
    prefix = 'cmf'
    attributes = (
        TypeAttribute('type'),
        WorkflowAttribute('workflow_history '),
        LocalRolesAttribute('local_role'))

    def getAttributeByName(self, name):

        for attr in self.attributes:
            if name in attr.getAttributeNames():
                return attr

    def processXml(self, context, node):
        return XmlNamespace.processXml(self, context, node)

    def getSchemaInfo(self):

        return [
            ("TypeInfo", "optional", TypeRNGSchemaFragment),
            ("SecurityInfo", "optional", SecurityRNGSchemaFragment),
            ("WorkflowInfo", "optional", WorkflowRNGSchemaFragment)]



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Persistent TALES Expression. Highly inspired (read copied)
from CMFCore's Expression.
"""

from Persistence import Persistent
from App.class_init import InitializeClass
from AccessControl import getSecurityManager, ClassSecurityInfo

from Products.PageTemplates.Expressions import getEngine
from Products.PageTemplates.Expressions import SecureModuleImporter


class Expression(Persistent):
    """A Persistent TALES Expression"""

    _text = ''
    _v_compiled = None

    security = ClassSecurityInfo()

    def __init__(self, text):
        self._text = text
        self._v_compiled = getEngine().compile(text)

    def __call__(self, econtext):
        compiled = self._v_compiled
        if compiled is None:
            compiled = self._v_compiled = getEngine().compile(self._text)
        res = compiled(econtext)
        if isinstance(res, Exception):
            raise res
        return res

InitializeClass(Expression)


def createExprContext(obj, **kw):
    """ Provides names for TALES expressions.
    """
    if obj is None:
        object_url = ''
    else:
        object_url = obj.absolute_url()

    user = getSecurityManager().getUser()

    data = {
        'object_url':   object_url,
        'object':       obj,
        'nothing':      None,
        'request':      getattr(obj, 'REQUEST', None),
        'modules':      SecureModuleImporter,
        'user':         user,
        }
    if 'mimetype' in kw and not 'content_type' in kw:
        # Alias content_type to mimetype
        kw['content_type'] = kw['mimetype']
    data.update(kw)
    for k in ('filename', 'content_type', 'data', 'mimetype'):
        data.setdefault(k, None)
    return getEngine().getContext(data)



"""
$Id$
"""

from cStringIO import StringIO
from Products.CMFCore.utils import getToolByName
from Products.Marshall import registry
from Products.Marshall.config import TOOL_ID as tool_id

add_registry = registry.manage_addRegistry

def install_tool(self, out):
    tool = getToolByName(self, tool_id, None)
    if tool is not None:
        out.write('Registry was already installed.\n')
        return
    add_registry(self)
    out.write('Registry installed sucessfully.\n')

def install(self, out=None):
    if out is None:
        out = StringIO()

    install_tool(self, out)

    return out.getvalue()



##################################################################
# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004 EnfoldSystems, LLC
# Copyright (C) 2004 ObjectRealms, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
from plone.uuid.interfaces import IUUID

"""
Serialize AT Schema Attributes

Created: 10/11/2004
Authors: Kapil Thangavelu <k_vertigo@objectrealms.net>
         Sidnei De Silva <sidnei@awkly.org>

$Id: $
"""
import re
from sets import Set

from Products.CMFCore.utils import getToolByName
from Products.Archetypes import config as atcfg
from Products.Archetypes.debug import log
from Products.Archetypes.interfaces import IBaseUnit
from Products.Archetypes.interfaces import IObjectField
from Products.Archetypes import public as atapi
from Products.Marshall import config
from Products.Marshall.handlers.atxml import XmlNamespace
from Products.Marshall.handlers.atxml import SchemaAttribute
from Products.Marshall.handlers.atxml import getRegisteredNamespaces
from Products.Marshall.exceptions import MarshallingException
from Products.Marshall import utils
from DateTime import DateTime
import transaction

_marker = object()


# Find control characters (0-31, 127) excluding:
#   9 (\t, tab)
#   10 (\n, new line character)
#   13 (\r, carriage return)
#   127 (\x1f, delete)
# which are handled properly by python xml libraries such
# as xml.dom.minidom and elementtree
CTRLCHARS_RE = re.compile(r'[\x00-\x08\x0b-\x0c\x0e-\x1f\x7f]')


def has_ctrlchars(value):
    if CTRLCHARS_RE.search(value):
        return True
    return False


class BoundReference(object):

    __slots__ = ('ns_data', 'attribute', 'instance')

    def __init__(self, ns_data, attribute, instance):
        self.ns_data = ns_data
        self.attribute = attribute
        self.instance = instance

    def resolve(self, context):
        self.attribute.deserialize(self.instance, self.ns_data)


class ATAttribute(SchemaAttribute):

    def get(self, instance):
        values = atapi.BaseObject.__getitem__(instance, self.name)
        if not isinstance(values, (list, tuple)):
            values = [values]
        return filter(None, values)

    def serialize(self, dom, parent_node, instance, options={}):

        encode_ctrlchars = options.get('encode_ctrlchars', True)

        values = self.get(instance)
        if not values:
            return

        for value in values:
            node = dom.createElementNS(self.namespace.xmlns, "field")
            name_attr = dom.createAttribute("name")
            name_attr.value = self.name
            node.setAttributeNode(name_attr)

            # try to get 'utf-8' encoded string
            if isinstance(value, unicode):
                value = value.encode('utf-8')
            elif IBaseUnit.providedBy(value):
                value = value.getRaw(encoding='utf-8')
            else:
                value = str(value)

            if self.isReference(instance):
                if config.HANDLE_REFS:
                    ref_node = dom.createElementNS(self.namespace.xmlns,
                                                    'reference')
                    uid_node = dom.createElementNS(self.namespace.xmlns,
                                                    'uid')
                    value = dom.createTextNode(value)
                    uid_node.append(value)
                    ref_node.append(uid_node)
                    node.append(ref_node)
            elif (encode_ctrlchars and
                  isinstance(value, str) and
                  has_ctrlchars(value)):
                value = value.encode('base64')
                attr = dom.createAttributeNS(self.namespace.xmlns,
                                             'transfer_encoding')
                attr.value = 'base64'
                node.setAttributeNode(attr)
                value_node = dom.createCDATASection(value)
                node.appendChild(value_node)
            else:
                value_node = dom.createTextNode(value)
                node.appendChild(value_node)

            # set the mimetype if it is available
            field = instance.schema._fields[self.name]
            if IObjectField.providedBy(field):
                mime_attr = dom.createAttribute('mimetype')
                mime_attr.value = field.getContentType(instance)
                node.setAttributeNode(mime_attr)

            node.normalize()
            parent_node.appendChild(node)

        return True

    def processXmlValue(self, context, value):
        if value is None:
            return

        value = value.strip()
        if not value:
            return

        # decode node value if needed
        te = context.node.get('transfer_encoding', None)
        if te is not None:
            value = value.decode(te)

        context_data = context.getDataFor(self.namespace.xmlns)
        data = context_data.setdefault(self.name, {'mimetype': None})
        mimetype = context.node.get('mimetype', None)
        if mimetype is not None:
            data['mimetype'] = mimetype

        if data.has_key('value'):
            svalues = data['value']
            if not isinstance(svalues, list):
                data['value'] = svalues = [svalues]
            svalues.append(value)
            return
        else:
            data['value'] = value

    def deserialize(self, instance, ns_data, options={}):
        if not ns_data:
            return
        data = ns_data.get( self.name )
        if data is None:
            return
        values = data.get('value', None)
        if not values:
            return

	# check if we are a schema attribute
        if self.isReference( instance ):
            values = self.resolveReferences( instance, values)
            if not config.HANDLE_REFS :
                return

        field = instance.Schema()[self.name]
        mutator = field.getMutator(instance)
        if not mutator:
            # read only field no mutator, but try to set value still
            # since it might reflect object state (like ATCriteria)
            field = instance.getField( self.name ).set( instance, values )
            #raise AttributeError("No Mutator for %s"%self.name)
            return
        if self.name == "id":
            transaction.savepoint()

        if instance.getField(self.name).type == 'datetime':
            # make sure DateTime fields are constructed properly
            # by explicitly constructing a DateTime instance
            mutator(DateTime(values))
        else:
            mutator(values)

        # set mimetype if possible
        mimetype = data.get('mimetype', None)
        if (mimetype is not None and IObjectField.providedBy(field)):
            field.setContentType(instance, mimetype)

    def resolveReferences(self, instance, values):
        ref_values = []
        for value in values:
            if not isinstance(value, Reference):
                ref_values.append(value)
                continue
            ref = value.resolve(instance)
            if ref is None: #just for dup behavior
                raise MarshallingException(
                    "Could not resolve reference %r" % value)
            ref_values.append(ref)
        return ref_values

    def isReference(self, instance):
        return not not isinstance(instance.Schema()[self.name],
                                  atapi.ReferenceField)


class ReferenceAttribute(SchemaAttribute):

    __slots__ = ('reference', 'name')

    def __init__(self, name, reference):
        super(ReferenceAttribute, self).__init__(name)
        self.reference = reference

    def processXml(self, context, node):
        return True

    def processXmlValue(self, context, value):
        self.reference[self.name] = value.strip()


class ArchetypeUID(SchemaAttribute):

    def serialize(self, dom, parent_node, instance, options={}):
        value = IUUID(instance, '')
        node = dom.createElementNS(Archetypes.xmlns, "uid")
        nvalue = dom.createTextNode(value)
        node.appendChild(nvalue)
        parent_node.appendChild(node)

    def deserialize(self, instance, ns_data):
        values = ns_data.get(self.name)
        if not values:
            return
        self.resolveUID(instance, values)

    def resolveUID(self, instance, values):
        assert not isinstance(values, (list, tuple))
        at_uid = values
        #existing = getattr(instance, atcfg.UUID_ATTR, _marker)
        existing = IUUID(instance, _marker)
        if existing is _marker or existing != at_uid:
            ref = Reference(uid=at_uid)
            target = ref.resolve(instance)
            if target is not None:
                raise MarshallingException(
                        "Trying to set uid of "
                        "%s to an already existing uid "
                        "clashed with %s" % (
                        instance.absolute_url(), target.absolute_url()))
            instance._setUID(at_uid)


RNGSchemaFragment = '''
  <define name="ArchetypesFields"
          ns="http://plone.org/ns/archetypes/"
          xmlns:note="atxml:annotations"
          datatypeLibrary="http://www.w3.org/2001/XMLSchema-datatypes"
          xmlns="http://relaxng.org/ns/structure/1.0">
    <element name="field">
      <note:info>
        All non-standard Archetypes fields are represented by a 'field'
        element, their id specified by an attribute 'id'.
      </note:info>
      <attribute name="id" />
      <choice>
        <text />
        <zeroOrMore>
          <element name="reference">
            <note:info>
              References can be made by UID (Archetypes),
              relative path, or by specifying a group of values
              that can uniquely identify a piece of content.
            </note:info>
            <choice>
              <zeroOrMore>
                <element name="uid"><text /></element>
              </zeroOrMore>
              <zeroOrMore>
                <element name="path"><text /></element>
              </zeroOrMore>
              <zeroOrMore>
                <ref name="Metadata" />
              </zeroOrMore>
            </choice>
          </element>
        </zeroOrMore>
      </choice>
    </element>
  </define>
  '''


class Archetypes(XmlNamespace):

    xmlns = config.AT_NS
    prefix = None
    attributes = []

    def __init__(self):
        super(Archetypes, self).__init__()
        self.last_schema_id = None
        self.in_reference_mode = False
        self.new_reference_p = True

        uid_attribute = ArchetypeUID('uid')
        uid_attribute.setNamespace(self)

        self.at_fields = {'uid': uid_attribute}

    def getAttributeByName(self, schema_name, context=None):
        if context is not None and schema_name not in self.at_fields:
            if not schema_name in context.instance.Schema():
                return
                raise AssertionError("invalid attribute %s" % (schema_name))

        if schema_name in self.at_fields:
            return self.at_fields[schema_name]

        attribute = ATAttribute(schema_name)
        attribute.setNamespace(self)

        return attribute

    def getAttributes(self, instance, exclude_attrs=()):

        # remove fields delegated to other namespaces
        fields = []
        for ns in getRegisteredNamespaces():
            if ns.uses_at_fields:
                fields.extend(ns.getATFields())
        assert len(Set(fields)) == len(fields), (
            "Multiple NS multiplexing field")

        field_keys = [k for k in instance.Schema().keys()
                      if k not in exclude_attrs and k not in fields]
        #Set(instance.Schema().keys())-mset

        # remove primary field if still present
        # XXX: we dont want to remove the PF, but want to be backward
        # XXX: compatible (how to do that best?)
        # p = instance.getPrimaryField()
        # pk = p and p.getName() or None
        # if pk and pk in field_keys:
        #     field_keys.remove( pk )

        for fk in field_keys:
            yield self.getAttributeByName(fk)

        # yield additional intrinsic at framework attrs
        for attribute in self.at_fields.values():
            yield attribute

    def serialize(self, dom, parent_node, instance, options):

        exclude_attrs = options.get('atns_exclude', ())

        for attribute in self.getAttributes(instance, exclude_attrs):
            if (hasattr(attribute, 'isReference') and
                attribute.isReference(instance)):
                continue
            attribute.serialize(dom, parent_node, instance, options)

    def deserialize(self, instance, ns_data, options):
        if not ns_data:
            return

        for attribute in self.getAttributes(instance):
            if (not config.HANDLE_REFS and
                hasattr(attribute, 'isReference') and
                attribute.isReference(instance)):
                # simply skip it then... Gogo
                continue
            attribute.deserialize(instance, ns_data)

    def processXml(self, context, data_node):

        tagname, namespace = utils.fixtag(data_node.tag, context.ns_map)

        if tagname == 'metadata':
            # ignore the container
            return False

        elif tagname == 'reference':
            # switch to reference mode, we tell the parser that we want
            # to explictly recieve all new node parse events, so we
            # can introspect the nested metadata that can be used
            # in reference specification.
            self.in_reference_mode = True
            self.new_reference_p = True
            assert self.last_schema_id
            context.setNamespaceDelegate(self)
            return False

        elif tagname == 'field':
            # basic at field specified, find the matching attribute
            # and annotate the data node with it
            schema_name = data_node.attrib.get('name', None)
            if schema_name is None:
                log("'id' attribute for at:field is deprecated, "
                    "use 'name' instead")
                schema_name = data_node.attrib.get('id')
            assert schema_name, "No field name specified in at:field element"
            #print "field", schema_name
            self.last_schema_id = schema_name
            attribute = self.getAttributeByName(schema_name, context)
            if attribute is None:
                #print "na", schema_name
                return False
            data_node.set('attribute', attribute)
            return True

        elif self.in_reference_mode:
            # if we get new metadata elements while in references, they
            # are stored as additional data for resolving the reference
            # latter.
            data = context.getDataFor(self.xmlns)
            srefs = data.setdefault(self.last_schema_id, [])

            # if we've already added a reference to the node data,
            # put additional reference specification data onto the
            # existing reference.
            if self.new_reference_p:
                ref = Reference()
                srefs.append(ref)
                self.new_reference_p = False
            else:
                ref = srefs[-1]

            attribute = ReferenceAttribute(data_node.name, ref)
            data_node.set('attribute', attribute)
            return True

        elif tagname in self.at_fields:
            # pseudo fields like uid which are specified in a custom manner
            attribute = self.getAttributeByName(tagname)
            if attribute is None:
                return False
            data_node.set('attribute', attribute)
            return True

        return False

    def processXmlEnd(self, name, context):
        if name == 'reference':
            context.setNamespaceDelegate(None)
            self.in_reference_mode = False
            self.last_schema_id = None # guard against bad xml

    def getSchemaInfo(self):
        return [("ArchetypesFields", "zeroOrMore", RNGSchemaFragment)]


class Reference(dict):

    index_map = dict([('title', 'Title'),
                      ('description', 'Description'),
                      ('creation_date', 'created'),
                      ('modification_data', 'modified'),
                      ('creators', 'Creator'),
                      ('subject', 'Subject'),
                      ('effectiveDate', 'effective'),
                      ('expirationDate', 'expires'),
                      ])

    def resolve(self, context):
        uid = self.get('uid')
        if uid is not None:
            rt = getToolByName(context, atcfg.REFERENCE_CATALOG)
            return rt.lookupObject(uid)
        path = self.get('path')
        if path is not None:
            return context.restrictedTraverse(path, None)
        catalog = getToolByName(context, 'portal_catalog')
        params = [(k, v) for k, v in self.items() \
                  if k not in ('uid', 'path')]
        kw = [(self.index_map.get(k), v) for k, v in params]
        kw = dict(filter(lambda x: x[0] is not None and x, kw))
        res = catalog(**kw)
        if not res:
            return None

        # First step: Try to filter by brain metadata
        # *Usually* a metadata item will exist with the same name
        # as the index.
        verify = lambda obj: filter(None, [obj[k] == v for k, v in kw.items()])
        for r in res:
            # Shortest path: If a match is found, return immediately
            # instead of checking all of the results.
            if verify(r):
                return r.getObject()

        # Second step: Try to get the real objects and look
        # into them. Should be *very* slow, so use with care.
        # We use __getitem__ to access the field raw data.
        verify = lambda obj: filter(None, [obj[k] == v for k, v in params])
        valid = filter(verify, [r.getObject() for r in res])
        if not valid:
            return None
        if len(valid) > 1:
            raise MarshallingException('Metadata reference does not '
                                       'uniquely identifies the reference.')
        return valid[0]



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id$
"""

import logging

from Products.CMFCore.utils import getToolByName
from Products.Archetypes.Marshall import Marshaller
from Products.Archetypes.utils import mapply
from Products.Marshall.config import logger
from Products.Marshall.config import TOOL_ID
from Products.Marshall.registry import getComponent
from Products.Marshall.exceptions import MarshallingException
from Acquisition import ImplicitAcquisitionWrapper


def getContext(obj, REQUEST=None):
    context = getattr(obj, 'aq_parent', None)
    if context is not None or REQUEST is None:
        return obj
    return REQUEST['PARENTS'][0]


class ControlledMarshaller(Marshaller):

    def __init__(self, fallback=None, demarshall_hook=None,
                 marshall_hook=None):
        Marshaller.__init__(self, demarshall_hook, marshall_hook)
        self.fallback = fallback

    def delegate(self, method, obj, data=None, file=None, **kw):
        if file is not None:
            kw['file'] = file
        __traceback_info__ = (method, obj, kw)
        context = getContext(obj, kw.get('REQUEST'))
        if context is not obj:
            # If the object is being created by means of a PUT
            # then it has no context, and some of the stuff
            # we are doing here may require a context.
            # Wrapping it in an ImplicitAcquisitionWrapper should
            # be safe as long as nothing tries to persist
            # a reference to the wrapped object.
            obj = ImplicitAcquisitionWrapper(obj, context)
        tool = getToolByName(obj, TOOL_ID, None)
        components = None
        if tool is not None:
            info = kw.copy()
            info['data'] = data
            info['mode'] = method
            components = tool.getMarshallersFor(obj, **info)
        else:
            # Couldn't find a context to get
            # hold of the tool or the tool is not installed.
            logger.log(logging.DEBUG, 'Could not find the marshaller tool. '
                'It might not be installed or you might not '
                'be providing enough context to find it.')
        # We just use the first component, if one is returned.
        if components:
            marshaller = getComponent(components[0])
        else:
            # If no default marshaller was provided then we complain.
            if self.fallback is None:
                raise MarshallingException(
                    "Couldn't get a marshaller for %r, %r" % (obj, kw))
            # Otherwise, use the default marshaller provided. Note it
            # must be an instance, not a factory.
            marshaller = self.fallback
        __traceback_info__ = (marshaller, method, obj, kw)
        args = (obj,)
        if method == 'demarshall':
            args += (data,)
        method = getattr(marshaller, method)
        return mapply(method, *args, **kw)

    def marshall(self, obj, **kw):
        if not 'data' in kw:
            kw['data'] = ''
        return self.delegate('marshall', obj, **kw)

    def demarshall(self, obj, data, **kw):
        if 'file' in kw:
            if not data:
                # XXX Yuck! Shouldn't read the whole file, never.
                # OTOH, if you care about large files, you should be
                # using the PrimaryFieldMarshaller or something
                # similar.
                data = kw['file'].read()
            del kw['file']

        return self.delegate('demarshall', obj, data, **kw)



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
"""

import re
import difflib

# Load fixture
from plone.app.testing.bbb import PloneTestCase


def normalize_tabs(s):
    s = re.sub(r"[ \t]+", " ", s)
    return s


def normalize_space(s):
    s =  re.sub(r"[\r\n]+", r'\r\n', s)
    return s


class BaseTest(PloneTestCase):
    """Base Test"""

    def compare(self, one, two):
        diff = difflib.ndiff(one.splitlines(), two.splitlines())
        diff = '\n'.join(list(diff))
        return diff

    def assertEqualsDiff(self, one, two, normalize=True):
        if normalize:
            one, two = normalize_tabs(one), normalize_tabs(two)
        one, two = normalize_space(one), normalize_space(two)
        self.failUnless(one.splitlines() == two.splitlines(),
                        self.compare(one, two))



from Products.Archetypes.public import *

class Person(BaseContent):

    schema = BaseSchema + Schema((
        ReferenceField('parents',
                       relationship='my_parents',
                       multiValued=True),
        ReferenceField('children',
                       relationship='my_children',
                       multiValued=True),
        ReferenceField('father',
                       relationship='my_father',
                       multiValued=False,
                       required=False),
        ReferenceField('mother',
                       relationship='my_mother',
                       multiValued=False,
                       required=False),
        LinesField('food_preference',
                   accessor='getFoodPrefs',
                   mutator='setFoodPrefs'),
        ))

registerType(Person, 'tests.Marshall')



##################################################################
# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004 EnfoldSystems, LLC
# Copyright (C) 2004 ObjectRealms, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

"""
$Id: $
"""

from DateTime import DateTime

from Products.Marshall.handlers.atxml import XmlNamespace
from Products.Marshall.handlers.atxml import SchemaAttribute


RNGSchemaFragment = '''
  <define name="DateInfo"
          xmlns:xmp="adobe:ns:meta"
          datatypeLibrary="http://www.w3.org/2001/XMLSchema-datatypes"
          xmlns="http://relaxng.org/ns/structure/1.0">
        <zeroOrMore>
          <element name="xmp:CreateDate"><data type="dateTime" /></element>
        </zeroOrMore>
        <zeroOrMore>
          <element name="xmp:ModifyDate"><data type="dateTime" /></element>
        </zeroOrMore>
  </define>
'''


class XMPDate(SchemaAttribute):

    def get(self, instance):
        return getattr(instance, self.field_id)

    def deserialize(self, instance, ns_data):
        value = ns_data.get(self.name)
        if not value:
            return
        value = DateTime(value)
        setattr(instance, self.field_id, value)

    def serialize(self, dom, parent_node, instance):
        value = self.get(instance)
        if isinstance(value, DateTime):
            value = value.HTML4()
        elname = "%s:%s" % (self.namespace.prefix, self.name)
        node = dom.createElementNS(self.namespace.xmlns,
                                   elname)
        value_node = dom.createTextNode(value)
        node.appendChild(value_node)
        node.normalize()
        parent_node.appendChild(node)


class AdobeXMP(XmlNamespace):

    xmlns = 'adobe:ns:meta'
    prefix = 'xmp'

    uses_at_fields = True

    attributes = (
        XMPDate('CreateDate', 'creation_date'),
        XMPDate('ModifyDate', 'modification_date'),)

    def getATFields(self):
        return ('creation_date', 'modification_date')

    def getSchemaInfo(self):
        return (("DateInfo", "zeroOrMore", RNGSchemaFragment),)



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
"""

from zope.interface import implementer

from OFS.OrderedFolder import OrderedFolder
from App.class_init import InitializeClass
from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import view

from Products.Marshall.interfaces import IPredicate
from Products.Marshall.interfaces import IMarshallRegistry
from Products.Marshall.config import TOOL_ID
from Products.Marshall.export import Export


class RegistryItem:

    def __init__(self, name, title, factory):
        self.name = name
        if not title:
            title = name
        self.title = title
        self.factory = factory

    def info(self):
        return {'name': self.name,
                'title': self.title}

    def create(self, id, title, expression, component_name):
        return self.factory(id, title or self.title, expression,
                            component_name)

comp_registry = {}
def registerComponent(name, title, component):
    comp_registry[name] = RegistryItem(name, title, component)

def getRegisteredComponents():
    return [item.info() for item in comp_registry.values()]

def getComponent(name):
    return comp_registry[name].factory()

registry = {}
def registerPredicate(name, title, component):
    component.predicate_type = title
    component.predicate_name = name
    registry[name] = RegistryItem(name, title, component)

def getRegisteredPredicates():
    return [item.info() for item in registry.values()]

def createPredicate(name, id, title, expression, component_name):
    return registry[name].create(id, title, expression, component_name)


@implementer(IMarshallRegistry)
class Registry(OrderedFolder, Export):
    """ A registry that holds predicates and applies them to
    objects in the hope of selecting the right one that matches
    a given set of constraints.
    """

    meta_type = 'Marshaller Registry'
    id = TOOL_ID
    security = ClassSecurityInfo()

    def __init__(self, id='', title=''):
        OrderedFolder.__init__(self, self.id)
        self.title = title or self.meta_type

    def all_meta_types(self):
        return OrderedFolder.all_meta_types(self, interfaces=(IPredicate,))

    security.declareProtected(view, 'getMarshallersFor')
    def getMarshallersFor(self, obj, **kw):
        set = []
        for predicate in self.objectValues():
            # An OrderedSet, in the lack of a real one.
            # The order in which predicates are evaluated
            # is managed by the underlying OrderedFolder impl.
            [set.append(s) for s in predicate.apply(obj, **kw)
             if s not in set]
        return tuple(set)

InitializeClass(Registry)


def manage_addRegistry(self, REQUEST=None, **ignored):
    """ Factory method that creates a Registry"""
    obj = Registry()
    self._setObject(obj.id, obj)

    if REQUEST is not None:
        REQUEST['RESPONSE'].redirect(self.absolute_url() + '/manage_main')
    else:
        return self._getOb(obj.id)



# bbb imports
import zope.deprecation
zope.deprecation.moved('Products.Archetypes.Marshall', 'Plone 5.1')

from Products.Archetypes.Marshall import NonLoweringMessage
from Products.Archetypes.Marshall import formatRFC822Headers
from Products.Archetypes.Marshall import parseRFC822
from Products.Archetypes.Marshall import RFC822Marshaller



def fixtag(tag, namespaces):
    """ returns the tag name seperated from the namespace uri
    as givein in {some.namespace}tag in elementtree"""

    # given a decorated tag (of the form {uri}tag), return prefixed
    # tag and namespace declaration, if any

    namespace_uri, tag = tag[1:].split("}", 1)
    ns = namespaces.get(namespace_uri)
    return tag, ns



import glob
import os
import re
import zipfile

# Load fixture
from Testing import ZopeTestCase
from Products.Marshall.tests.base import BaseTest

# Install our product
ZopeTestCase.installProduct('Marshall')
ZopeTestCase.installProduct('Archetypes')
ZopeTestCase.installProduct('ATContentTypes')

from Products.CMFCore.utils import getToolByName
from Products.Marshall import registry
from Products.Marshall.registry import Registry, getRegisteredComponents
from Products.Marshall.registry import getComponent
from Products.Marshall.tests import PACKAGE_HOME
tool_id = Registry.id

def normalize_xml(s):
    s = re.sub(r"[ \t]+", " ", s)
    return s


def import_file(relparts, fname, target, handler):
    marshaller = getComponent(handler)
    f = open(fname, 'rb+')
    content = f.read()
    f.close()
    curr = parent = target
    for p in relparts[:-1]:
        curr = parent.restrictedTraverse(p, None)
        if curr is None:
            parent.invokeFactory('Folder', p)
            curr = parent.restrictedTraverse(p)
        parent = curr
    obj_id = relparts[-1]
    obj = parent.restrictedTraverse(obj_id, None)
    if obj is None:
        parent.invokeFactory('Document', obj_id)
        obj = parent.restrictedTraverse(obj_id)
    marshaller.demarshall(obj, content)
    return

IGNORE_NAMES = ('CVS', '.svn')
def fromFS(base, target, metadata='atxml', data='primary_field'):
    paths = []
    ignore = lambda x: filter(None, [x.endswith(n) for n in IGNORE_NAMES])
    def import_metadata(relparts, fname, target, handler=metadata):
        return import_file(relparts, fname, target, handler)
    def import_data(relparts, fname, target, handler=data):
        return import_file(relparts, fname, target, handler)
    def import_func(arg, dirname, names):
        # Remove ignored filenames
        [names.remove(n) for n in names if ignore(n)]
        names = map(os.path.normcase, names)
        for name in names:
            fullname = os.path.join(dirname, name)
            if not os.path.isfile(fullname):
                continue
            fparts = fullname.split(os.sep)
            bparts = base.split(os.sep)
            relparts = fparts[len(bparts):]
            relpath = '/'.join(relparts)
            arg.append(relpath)
            if '.metadata' in dirname:
                relparts.remove('.metadata')
                import_metadata(relparts, fullname, target)
            else:
                import_data(relparts, fullname, target)
    os.path.walk(base, import_func, paths)
    return paths

class ExportTest(BaseTest):

    def afterSetUp(self):
        super(ExportTest, self).afterSetUp()
        self.loginAsPortalOwner()
        registry.manage_addRegistry(self.portal)
        self.tool = getToolByName(self.portal, tool_id)

    def test_export(self):
        self.portal.invokeFactory('Folder', 'test_data')
        self.folder = self.portal.test_data
        paths = fromFS(self.base, self.folder)
        paths.sort()
        obj_paths = filter(lambda x: '.metadata' not in x, paths)
        data = self.tool.export(self.folder, obj_paths)
        zipf = zipfile.ZipFile(data)
        self.assertEquals(zipf.testzip(), None)
        zipl = zipf.namelist()
        zipl.sort()
        self.assertEquals(zipl, paths)

def test_suite():
    import unittest
    suite = unittest.TestSuite()
    dirs = glob.glob(os.path.join(PACKAGE_HOME, 'export', '*'))
    comps = [i['name'] for i in getRegisteredComponents()]
    for d in dirs:
        prefix = os.path.basename(d)
        if prefix not in comps:
            continue
        k_dict = {'base':d,
                  'prefix':prefix}
        klass = type('%sExportTest' % prefix,
                     (ExportTest,),
                     k_dict)
        suite.addTest(unittest.makeSuite(klass))
    return suite



from Products.Marshall.handlers.atxml import registerNamespace

from adobens import AdobeXMP
from atns import Archetypes
from dcns import DublinCore
from cmfns import CMF

registerNamespace(DublinCore)
registerNamespace(AdobeXMP)
registerNamespace(Archetypes)
registerNamespace(CMF)

# we do cmf last because workflow might other get reset
# by a set id (manage_afterAdd hook) value change




# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id$
"""

from zope.interface import implementer

from OFS.SimpleItem import SimpleItem
from App.class_init import InitializeClass
from AccessControl import ClassSecurityInfo
from AccessControl.Permissions import view, manage_properties
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.Marshall.expression import Expression, createExprContext
from Products.Marshall.registry import createPredicate, registerPredicate
from Products.Marshall.registry import getRegisteredPredicates
from Products.Marshall.registry import getRegisteredComponents
from Products.Marshall.interfaces import IPredicate


@implementer(IPredicate)
class Predicate(SimpleItem):
    """ A Predicate for selecting marshallers.

    Each Predicate is composed of a Expression  which
    is evaluated in the context of the object to
    decide if the Predicate applies, and returns
    a component name to be used for that object.
    """

    meta_type = "Marshaller Predicate"
    predicate_type = None
    predicate_name = None
    security = ClassSecurityInfo()

    manage_options = (
        {'label': 'Edit', 'action': 'manage_changePredicateForm'},
        ) + SimpleItem.manage_options

    security.declareProtected('View management screens',
      'manage_changePredicateForm', 'manage_main', 'manage')

    manage_changePredicateForm = PageTemplateFile(
        '../www/predicateChange', globals(),
        __name__='manage_changePredicateForm')

    manage_changePredicateForm._owner = None
    manage = manage_main = manage_changePredicateForm

    def __init__(self, id, title, expression='', component_name=''):
        self.id = id
        self.title = title
        self.setExpression(expression)
        self.setComponentName(component_name)

    security.declareProtected(view, 'getPredicateType')
    def getPredicateType(self):
        return self.predicate_type

    security.declareProtected(view, 'getPredicateName')
    def getPredicateName(self):
        return self.predicate_name

    security.declareProtected(view, 'apply')
    def apply(self, obj, **kw):
        """ Evaluate expression using the object as
        context and return a component name if applicable.
        """
        if not self.getExpression():
            # Short-circuit for speedup when no expression.
            return (self.getComponentName(),)
        context = createExprContext(obj, **kw)
        if self.expression(context):
            return (self.getComponentName(),)
        return ()

    security.declareProtected(view, 'getComponentName')
    def getComponentName(self):
        """ Return the component name configured for
        this predicate.
        """
        return self._component_name

    security.declareProtected(manage_properties, 'setComponentName')
    def setComponentName(self, component_name):
        """ Change the component name """
        valid_components = [i['name'] for i in getRegisteredComponents()]
        if component_name not in valid_components:
            raise ValueError('Not a valid registered '
                             'component: %s' % component_name)
        self._component_name = component_name

    security.declareProtected(manage_properties, 'setExpression')
    def setExpression(self, expression):
        """ Change the expression """
        self._expression_text = expression
        self._expression = Expression(expression)

    security.declareProtected(view, 'expression')
    def expression(self, context):
        """ Evaluate the expression using context """
        return self._expression(context)

    security.declareProtected(view, 'getExpression')
    def getExpression(self):
        """ Get the expression as text """
        return self._expression_text

    def manage_availableComponents(self):
        return getRegisteredComponents()

    security.declareProtected(manage_properties, 'manage_changePredicate')
    def manage_changePredicate(self, title=None,
                               expression=None,
                               component_name=None,
                               REQUEST=None):
        """ Change the settings of this predicate """

        if title is not None:
            self.title = title
        if expression is not None:
            self.setExpression(expression)
        if component_name is not None:
            self.setComponentName(component_name)

        if REQUEST is not None:
            message = 'Predicate Constraints Changed.'
            return self.manage_changePredicateForm(
                manage_tabs_message=message,
                management_view='Edit')
        return self

InitializeClass(Predicate)
registerPredicate('default', 'Default Predicate', Predicate)


def manage_addPredicate(self, id, title, predicate, expression,
                        component_name, REQUEST=None):
    """ Factory method that creates a Property Set Predicate"""
    obj = createPredicate(predicate, id, title, expression, component_name)
    self._setObject(id, obj)

    if REQUEST is not None:
        next = 'manage_main'
        if hasattr(obj, 'manage_changeSettingsForm'):
            next = 'manage_changeSettingsForm'
        REQUEST['RESPONSE'].redirect('/'.join((self.absolute_url(), id, next)))

    return self._getOb(id)


def manage_availablePredicates(self):
    return getRegisteredPredicates()


def manage_availableComponents(self):
    return getRegisteredComponents()

manage_addPredicateForm = PageTemplateFile(
    '../www/predicateAdd', globals(),
    __name__='manage_addPredicateForm')

constructors = (
  ('manage_addPredicateForm', manage_addPredicateForm),
  ('manage_addPredicate', manage_addPredicate),
  ('manage_availablePredicates', manage_availablePredicates),
  ('manage_availableComponents', manage_availableComponents),
)



#!/usr/bin/env python
# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""
ATXML data uploader
Give it a list of ATXML files, and it will do a PUT
to the server of those files, plus any "content" with the
same base name but different extension. If no list is specified,
will use all in the current directory.

Doesn't currently iterate over directories.

If you want to use this for content without ATXML files,
set the ATXML extension as the content

"""

import os, sys, time
from optparse import OptionParser
from WebDAV import davlib     # not the standard davlib
from xml.dom import minidom

class FakeResponse: pass

def upload(content, conn, path):
    if content:
        return conn.put(path, content, content_type='text/html')
    else:
        x = FakeResponse()
        x.status = 0
        return x

def getText(nodelist):
    """DOM helper, for extracting text"""
    rc = ""
    for node in nodelist:
        if node.nodeType == node.TEXT_NODE:
            rc = rc + node.data
    return rc

start = time.time()

usage = "usage: %prog [options] server[:port] /path/to/col/ [atxml files...]"
parser = OptionParser(usage=usage)
parser.add_option("-u", "--username", dest="username",
                 help="WebDAV user name", metavar="USER",
                 default="admin")
parser.add_option("-p", "--password", dest="password",
                 help="WebDAV password", metavar="PASS",
                 default="admin")
parser.add_option("-a", "--atxml", dest="atxmlend",
                 help="File extension of ATXML", metavar=".atxml",
                 default=".atxml")
parser.add_option("-c", "--content", dest="contentend",
                 help="File extension of content", metavar=".xml",
                 default=".xml")
parser.add_option("-s", "--server", dest="serverend",
                 help="File extension on server", metavar=".html",
                 default=".html")
(options, args) = parser.parse_args()

if len(args) < 2:
    print "Error: Server and path are required. Use -h for help."
    print
    sys.exit(-1)

if len(args) > 2:
    print "Using ATXML files supplied on commandline:"
    files = args[2:]
else:
    print "Using all ATXML files in directory:"
    files = [x for x in os.listdir(".") if x.endswith(options.atxmlend)]

server = args[0]
pathstarts = args[1].rstrip('/')+'/'
conn = davlib.DAV(server, timeout=300)
conn.setauth(options.username, options.password)

bad = []
count = 0
for file in files:
    try:
        metafile = open(file)
        metacontent = metafile.read()

        name = metafile.name
        print name,
        if not name.endswith(options.atxmlend):
            print "Not properly named. Expected to end with '%s'" % options.atxmlend
            continue
        if len(options.atxmlend) > 0:
            id = fid = name[:-len(options.atxmlend)]
        else:
            id = fid = name

        metafile.close()

        try:
            f = open(fid+options.contentend)
            content = f.read()
            f.close()
        except IOError:
            content = ""

        # find our true id, regardless of file name, if specified in the file
        try:
            dom = minidom.parseString(metacontent)
            field = [x for x in dom.getElementsByTagName('field') \
                     if x.getAttribute("id")=="id"][0]
            id = getText(field.childNodes) or fid
            dom.unlink()
        except:
            pass

        path = "%s%s%s" % (pathstarts,id,options.serverend)
        print "-> %s" % path,

        errormeta = None
        response = upload(metacontent, conn, path)
        if not response.status in(201, 204):
            errormeta = "%s %s" % (response.status,response.reason)
        #print response.status,

        errorcontent = None
        response = upload(content, conn, path)
        if not response.status in (204, 0):
            errorcontent = "%s %s" % (response.status,response.reason)
        #print response.status,

        if errormeta or errorcontent:
            print "FAIL"
            print "   Metadata: %s" % errormeta
            print "   Content:  %s" % errorcontent
            bad += [name]
        else:
            print "ok"
            count += 1
    except "ExpatError":
        print "FAIL: malformed ATXML file"
        bad += [name]
    except Exception, e:
        print "FAIL"
        print "  %s" % e
        bad += [name]

conn.close()

end = time.time()

print "Successfully imported: %s pieces out of %s" % (`count`,`len(files)`)
print "Time elapsed: %s" % `end-start`
print
print "There were %s failures: " % `len(bad)`,
for num in bad:
    print num,



import os
import doctest

# Load fixture
from Products.Marshall.tests.base import BaseTest

from Products.CMFCore.utils import getToolByName
from Products.Marshall import registry
from Products.Marshall.registry import Registry
from Products.Marshall.registry import getComponent
from Products.Marshall.tests import PACKAGE_HOME
from Products.Marshall.tests.examples import blob

from plone.app.testing.bbb import PTC_FUNCTIONAL_TESTING
from plone.testing import layered


tool_id = Registry.id


def get_data(fname):
    return open(os.path.join(PACKAGE_HOME, 'data', fname), 'rb').read()


class BlobMarshallTest(BaseTest):

    def afterSetUp(self):
        super(BlobMarshallTest, self).afterSetUp()
        self.loginAsPortalOwner()
        registry.manage_addRegistry(self.portal)
        self.tool = getToolByName(self.portal, tool_id)
        self.marshaller = getComponent('atxml')

    def createBlob(self, ctx, id, **kw):
        blob.addBlob(ctx, id, **kw)
        ob = ctx._getOb(id)
        ob.indexObject()
        return ob

    def _test_blob_roundtrip(self, fname, data, mimetype, filename):
        blob = self.createBlob(self.portal, 'blob')

        field = blob.Schema()[fname]
        field.set(blob, data, mimetype=mimetype, filename=filename)

        # Marshall to XML
        ctype, length, got = self.marshaller.marshall(blob)

        # Populate from XML
        self.marshaller.demarshall(blob, got)

        # Marshall from XML again and compare to see if it's
        # unchanged.
        ctype, length, got2 = self.marshaller.marshall(blob)
        self.assertEqualsDiff(got, got2)

    def x_test_blob_image(self):
        # fails with IOError: cannot identify image file <cStringIO.StringI object at 0x7f07776de2d8>
        data = get_data('image.gif')
        self._test_blob_roundtrip('aimage', data, 'image/gif', 'image.gif')

    def test_blob_file_text(self):
        data = get_data('file.txt')
        self._test_blob_roundtrip('afile', data, 'text/plain', 'file.txt')

    def test_blob_file_binary(self):
        data = get_data('file.pdf')
        self._test_blob_roundtrip('afile', data, 'application/pdf', 'file.pdf')

    def test_blob_file_html(self):
        data = get_data('file.html')
        self._test_blob_roundtrip('afile', data, 'text/html', 'file.html')

    def test_blob_text_text(self):
        data = get_data('file.txt')
        self._test_blob_roundtrip('atext', data, 'text/plain', 'file.txt')

    def test_blob_text_binary(self):
        data = get_data('file.pdf')
        self._test_blob_roundtrip('atext', data, 'application/pdf', 'file.pdf')

    def test_blob_text_html(self):
        data = get_data('file.html')
        self._test_blob_roundtrip('atext', data, 'text/html', 'file.html')


OPTIONFLAGS = (doctest.ELLIPSIS |
               doctest.NORMALIZE_WHITESPACE |
               doctest.REPORT_ONLY_FIRST_FAILURE)


def test_suite():
    import unittest
    suite = unittest.TestSuite()

    suite.addTest(layered(doctest.DocFileSuite(
        'doc/README.txt',
        package='Products.Marshall',
        optionflags=OPTIONFLAGS),
        layer=PTC_FUNCTIONAL_TESTING))
    suite.addTest(unittest.makeSuite(BlobMarshallTest))
    return suite



from Products.CMFCore.utils import getToolByName

def export(self):
    ct = getToolByName(self, 'portal_catalog')
    ut = getToolByName(self, 'portal_url')
    mt = getToolByName(self, 'marshaller_registry')
    paths = []
    for pt in ('News Item', 'Link', 'Document', 'Image', 'File'):
        for r in ct(portal_type=pt):
            paths.append(ut.getRelativeUrl(r.getObject()))

    response = self.REQUEST.RESPONSE
    response.setHeader('content-type', 'application/octet-stream')
    response.setHeader('content-disposition', 'attachment; filename=export.zip')
    data = mt.export(self, paths).read()
    response.setHeader('content-length', len(data))
    return response.write(data)



# Load fixture
from Testing import ZopeTestCase
from Products.Marshall.tests.base import BaseTest

# Install our product
ZopeTestCase.installProduct('Marshall')
ZopeTestCase.installProduct('Archetypes')

from Products.CMFCore.utils import getToolByName
from Products.Marshall.predicates import add_predicate
from Products.Marshall.config import TOOL_ID as tool_id
from Products.Marshall import config
from Products.Marshall import registry


class PredicateTest(BaseTest):

    def afterSetUp(self):
        super(PredicateTest, self).afterSetUp()
        self.loginAsPortalOwner()
        registry.manage_addRegistry(self.portal)
        self.tool = getToolByName(self.portal, tool_id)

    def get(self, obj, **kw):
        return self.tool.getMarshallersFor(obj, **kw)

    def reverse(self):
        ids = self.tool.objectIds()
        for pos, id in enumerate(ids):
            self.tool.moveObjectToPosition(id, len(ids) - pos)


class DefaultPredicateTest(PredicateTest):

    def test_expression_filename(self):
        add_predicate(self.tool, id='doc',
                      title='.doc extension',
                      predicate='default',
                      expression="python: filename and filename.endswith('.doc')",
                      component_name='primary_field')
        add_predicate(self.tool, id='txt',
                      title='.txt extension',
                      predicate='default',
                      expression="python: filename and filename.endswith('.txt')",
                      component_name='rfc822')
        add_predicate(self.tool, id='default',
                      title='default',
                      predicate='default',
                      expression='',
                      component_name='primary_field')

        self.assertEquals(
            self.get(None, filename='sample.doc'), ('primary_field',))
        self.assertEquals(
            self.get(None, filename='sample.txt'),
            ('rfc822', 'primary_field'))

        # Make sure reordering the predicates does change the order
        # in which ids are returned.
        self.reverse()
        self.assertEquals(
            self.get(None, filename='sample.txt'),
            ('primary_field', 'rfc822'))

        self.assertEquals(self.get(None, filename='sample.rot'),
                          ('primary_field',))
        # Make sure it works even if no filename kw is passed in
        self.assertEquals(self.get(None), ('primary_field',))

    if config.hasLibxml2:
        def test_expression_content_type(self):
            add_predicate(
                self.tool, id='text_xml',
                title='text/xml content type',
                predicate='default',
                expression="python: content_type and content_type == 'text/xml'",
                component_name='simple_xml')
            add_predicate(
                self.tool,
                id='text_plain',
                title='text/plain content type',
                predicate='default',
                expression="python: content_type and content_type == 'text/plain'",
                component_name='rfc822')
            add_predicate(
                self.tool,
                id='default',
                title='default',
                predicate='default',
                expression="",
                component_name='primary_field')

            self.assertEquals(
                self.get(None, content_type='text/xml'),
                ('simple_xml', 'primary_field'))
            self.assertEquals(self.get(None, content_type='text/plain'),
                              ('rfc822', 'primary_field'))

            # Make sure reordering the predicates does change the order
            # in which ids are returned.
            self.reverse()
            self.assertEquals(self.get(None, content_type='text/plain'),
                              ('primary_field', 'rfc822'))

            # Make sure it works even if no content_type kw is passed in
            self.assertEquals(self.get(None), ('primary_field',))

    def test_expression_data(self):
        add_predicate(
            self.tool,
            id='data_len_3',
            title='data length > 3',
            predicate='default',
            expression="python: data and len(data) > 3",
            component_name='primary_field')
        add_predicate(
            self.tool,
            id='data_len_2',
            title='data length == 2',
            predicate='default',
            expression="python: data and len(data) == 2",
            component_name='rfc822')
        add_predicate(
            self.tool,
            id='default',
            title='default',
            predicate='default',
            expression='',
            component_name='primary_field')
        self.assertEquals(self.get(None, data='4242'), ('primary_field',))
        self.assertEquals(self.get(None, data='42'), ('rfc822', 'primary_field'))

        # Make sure reordering the predicates does change the order
        # in which ids are returned.
        self.reverse()
        self.assertEquals(self.get(None, data='42'), ('primary_field', 'rfc822'))

        # Make sure it works even if no data kw is passed in
        self.assertEquals(self.get(None), ('primary_field',))


class XMLNSPredicateTest(PredicateTest):

    def test_xmlns_element(self):
        p = add_predicate(
            self.tool,
            id='xmlns_attr',
            title='xmlns',
            predicate='xmlns_attr',
            expression='',
            component_name='primary_field')
        p.edit(element_name='test', value='FooBar')
        data = """<?xml version="1.0" encoding="UTF-8"?>
        <test>
           FooBar
        </test>
        """
        bad_data = """<?xml version="1.0" encoding="UTF-8"?>
        <test>
           FooBaz
        </test>
        """
        self.assertEquals(self.get(None, data=data), ('primary_field',))
        self.assertEquals(self.get(None, data=bad_data), ())
        self.assertEquals(self.get(None, data=''), ())

    def test_xmlns_element_ns(self):
        p = add_predicate(
            self.tool,
            id='xmlns_attr',
            title='xmlns',
            predicate='xmlns_attr',
            expression='',
            component_name='primary_field')
        p.edit(element_ns='http://foo.com/bar', element_name='test',
               value='FooBar')
        data = """<?xml version="1.0" encoding="UTF-8"?>
        <t:test xmlns:t="http://foo.com/bar">
                FooBar
        </t:test>
        """
        self.assertEquals(self.get(None, data=data), ('primary_field',))

    def test_xmlns_element_ns_att(self):
        p = add_predicate(
            self.tool,
            id='xmlns_attr',
            title='xmlns',
            predicate='xmlns_attr',
            expression='',
            component_name='primary_field')
        p.edit(element_ns='http://foo.com/bar', element_name='test',
               attr_name='wot', value='FooBar')
        data = """<?xml version="1.0" encoding="UTF-8"?>
        <t:test xmlns:t="http://foo.com/bar"
           wot="FooBar" />
        """
        self.assertEquals(self.get(None, data=data), ('primary_field',))

    def test_xmlns_element_ns_att_ns(self):
        p = add_predicate(
            self.tool,
            id='xmlns_attr',
            title='xmlns',
            predicate='xmlns_attr',
            expression='',
            component_name='primary_field')
        p.edit(element_ns='http://foo.com/bar', element_name='test',
               attr_ns='http://foo.com/bar', attr_name='wot',
               value='FooBar')
        data = """<?xml version="1.0" encoding="UTF-8"?>
        <t:test xmlns:t="http://foo.com/bar"
           t:wot="FooBar" />
        """
        self.assertEquals(self.get(None, data=data), ('primary_field',))


def test_suite():
    import unittest
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(DefaultPredicateTest))
    suite.addTest(unittest.makeSuite(XMLNSPredicateTest))
    return suite



##################################################################
# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004 EnfoldSystems, LLC
# Copyright (C) 2004 ObjectRealms, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

"""
"""

# Georg Gogo. BERNHARD: UUencoded pluggable namespace for all at fields

#from DateTime import DateTime
from Products.Marshall.public import XmlNamespace, SchemaAttribute
from cStringIO import StringIO

import base64
import uu

DEBUG = False


class UUAttribute(SchemaAttribute):

#    def get(self, instance):
#        values = atapi.BaseObject.__getitem__(instance, self.name)
#
#        import pdb; pdb.set_trace() # @@@
#
#        field = instance.Schema().getField(self.name)
#        values = field.getRaw(instance)
#
#
#
#
#        if not isinstance(values, (list, tuple)):
#            values = [values]
#        return filter(None, values)

#    def get(self, instance):
#        values = atapi.BaseObject.__getitem__(instance, self.name)
#        if not isinstance(values, (list, tuple)):
#            values = [values]
#        return filter(None, values)

    def base64encode(self, string):
        return base64.encodestring(string)

    def base64decode(self, string):
        return base64.decodestring(string)

    def uuencode(self, string):
        # uu encode a srting
        stringio = StringIO()
        uu.encode(StringIO(string), stringio)
        return stringio.getvalue()

    def uudecode(self, string):
        # uu decode a srting
        stringio = StringIO()
        uu.decode(StringIO(string), stringio)
        return stringio.getvalue()

    def char_encode(self, string):
        return self.uuencode(string)

    def char_decode(self, string):
        return self.uudecode(string)

    def deserialize(self, instance, ns_data):
        values = ns_data.get(self.name)
        if not values:
            return

        field = instance.Schema().getField(self.name)
        baseunit = field.getBaseUnit(instance)
        isBinary = baseunit.isBinary()
        if isBinary:
            values = self.char_decode(values)

        mutator = instance.Schema()[self.name].getMutator(instance)
        if not mutator:
            #raise AttributeError("No Mutator for %s"%self.name)
            return

        mutator(values)

    def serialize(self, dom, parent_node, instance):

        if DEBUG:
            print "uuns/UUattribute", instance, self.name

        field = instance.Schema().getField(self.name)
        baseunit = field.getBaseUnit(instance)
        value = baseunit.getRaw(encoding=baseunit.original_encoding)

        isBinary = baseunit.isBinary()
        if isBinary:
            value = self.char_encode(value)

        elname = "%s:%s" % (self.namespace.prefix, self.name)
        node = dom.createElementNS(self.namespace.xmlns,
                                   elname)
        value_node = dom.createTextNode(value)
        node.appendChild(value_node)
        node.normalize()

        if  isBinary:
            id_attr = dom.createAttributeNS(self.namespace.xmlns, "binary")
            id_attr.value = "uuencoded"
            node.setAttributeNode(id_attr)

        parent_node.appendChild(node)


class UUNS(XmlNamespace):

    xmlns = 'uuns:ns:data'
    prefix = 'at_data'
    attributes = []

    #uses_at_fields = True

    def getAttributeByName(self, schema_name, context=None):
        if context is not None and schema_name not in self.at_fields:
            assert schema_name in context.instance.Schema()

        #if schema_name in self.at_fields:
        #    return self.at_fields[ schema_name ]

        attribute = UUAttribute(schema_name)
        attribute.setNamespace(self)

        return attribute

    def getAttributes(self, instance):

        field_keys = instance.Schema().keys()
        if DEBUG:
            print "UUNS/getAttributes", field_keys
        #import pdb; pdb.set_trace() # @@@
        #return fields

        # remove fields delegated to other namespaces
        #fields = []
        #for ns in getRegisteredNamespaces():
        #    if ns.uses_at_fields:
        #        fields.extend(ns.getATFields())
        #mset = Set(fields)
        #assert len(mset) == len(fields), "Multiple NS multiplexing field"
        #
        #field_keys = Set(instance.Schema().keys())-mset
        #
        ## remove primary field if still present
        #p = instance.getPrimaryField()
        #pk = p and p.getName() or None
        #if pk:
        #    field_keys.remove(pk)
        #

#        import pdb; pdb.set_trace() # @@@

        for fk in field_keys:
            field = instance.Schema().getField(fk)

            if hasattr(field, 'getBaseUnit'):
                baseunit = field.getBaseUnit(instance)
            else:
                #print ">>>>>>>> field==", field, "no getBaseUnit!"
                continue

            isBinary = baseunit.isBinary()

            if isBinary or instance.getPrimaryField() == field:
                yield self.getAttributeByName(fk)

            if DEBUG:
                print fk, isBinary

        #
        ## yield additional intrinsic at framework attrs
        #for attribute in self.at_fields.values():
        #    yield attribute

    def serialize(self, dom, parent_node, instance):
        for attribute in self.getAttributes(instance):
            attribute.serialize(dom, parent_node, instance)

    def deserialize(self, instance, ns_data):
        if not ns_data:
            return

        for attribute in self.getAttributes(instance):
            attribute.deserialize(instance, ns_data)



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2008 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id$
"""

import pkg_resources

# Kick off Extensions.Install import
from Products.Marshall.Extensions import Install
del Install

# Kick off handler registration
from Products.Marshall import handlers

# Kick off namespace registration
from Products.Marshall import namespaces

from Products.Marshall.marshaller import ControlledMarshaller


def initialize(context):
    from Products.Marshall import registry
    from Products.Marshall import predicates

    context.registerClass(
        registry.Registry,
        permission='Add Marshaller Registry',
        constructors=(registry.manage_addRegistry,),
        icon='www/registry.png',
        )

    context.registerClass(
        instance_class=predicates.Predicate,
        permission='Add Marshaller Predicate',
        constructors=predicates.constructors,
        icon='www/registry.png')

    # Load marshallers registered by the way of egg entry points.
    for ep in pkg_resources.iter_entry_points('marshall.component'):
        component = ep.load()
        registry.registerComponent(
            ep.name, getattr(component, '__name__', ep.name), component)



from Products.Archetypes.public import *

class Blob(BaseContent):

    schema = BaseSchema + Schema((
        StringField('astring'),
        FileField('afile'),
        TextField('atext'),
        ImageField('aimage'),
        ))

registerType(Blob, 'tests.Marshall')



from setuptools import setup, find_packages

version = '2.2.1'

setup(
    name='Products.Marshall',
    version=version,
    description="Configurable Marshallers for Archetypes",
    long_description=(open("README.rst").read() + "\n" +
                      open("CHANGES.rst").read()),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Zope2",
        "Framework :: Plone",
        "Framework :: Plone :: 5.0",
        "Framework :: Plone :: 5.1",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: File Transfer Protocol (FTP)",
    ],
    keywords='web zope application server webdav ftp',
    license="GPL",
    author='Sidnei da Silve and others',
    author_email='plone-developers@lists.sourceforge.net',
    url='https://pypi.python.org/pypi/Products.Marshall',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['Products'],
    include_package_data=True,
    zip_safe=False,
    extras_require=dict(
      test=[
          'Products.ATContentTypes',
      ]
    ),
    install_requires=[
        'setuptools',
        'transaction',
        'plone.uuid',
        'zope.contenttype',
        'zope.deprecation',
        'zope.interface',
        'Products.Archetypes',
        'Products.CMFCore',
        'Products.GenericSetup',
        'Acquisition',
        'DateTime',
        'ExtensionClass',
        'Zope2',
    ],
    )



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
"""

import os
import zipfile
from cStringIO import StringIO
from App.class_init import InitializeClass
from ExtensionClass import Base
from AccessControl import ClassSecurityInfo
from Acquisition import aq_base
from Products.CMFCore.permissions import ManagePortal
from Products.PageTemplates.PageTemplateFile import PageTemplateFile


class Export(Base):

    security = ClassSecurityInfo()

    security.declareProtected(ManagePortal, 'atxml_template')
    atxml_template = PageTemplateFile('www/atxml', globals(),
                                      __name__='atxml_template')

    security.declareProtected(ManagePortal, 'marshall_data')
    def marshall_data(self, obj):
        from Products.Marshall.registry import getComponent
        marshaller = getComponent('primary_field')
        return self.marshall(obj, marshaller)

    security.declareProtected(ManagePortal, 'marshall_metadata')
    def marshall_metadata(self, obj):
        from Products.Marshall.registry import getComponent
        marshaller = getComponent('atxml')
        return self.marshall(obj, marshaller)

    security.declareProtected(ManagePortal, 'marshall')
    def marshall(self, obj, marshaller):
        REQUEST = obj.REQUEST
        RESPONSE = REQUEST.RESPONSE
        ddata = marshaller.marshall(obj, REQUEST=REQUEST,
                                    RESPONSE=RESPONSE)
        if hasattr(aq_base(obj), 'marshall_hook') \
           and obj.marshall_hook:
            ddata = obj.marshall_hook(ddata)

        content_type, length, data = ddata

        if type(data) is type(''):
            return StringIO(data)

        s = StringIO()
        while data is not None:
            s.write(data.data)
            data = data.next
        s.seek(0)
        return s

    security.declareProtected(ManagePortal, 'export')
    def export(self, context, paths):
        data = StringIO()
        out = zipfile.ZipFile(data, 'w')

        for path in paths:
            filename = os.path.basename(path)
            dir_path = os.path.dirname(path)
            obj = context.restrictedTraverse(path)

            # Write data
            fpath = os.path.join(dir_path, filename)
            stream = self.marshall_data(obj)
            out.writestr(fpath, stream.read())

            # Write metadata
            metadata_path = os.path.join(dir_path, '.metadata')
            fpath = os.path.join(metadata_path, filename)
            stream = self.marshall_metadata(obj)
            out.writestr(fpath, stream.read())
        out.close()
        data.seek(0)
        return data

    security.declareProtected(ManagePortal, 'export_info')
    def export_info(self, context, info):
        data = StringIO()
        out = zipfile.ZipFile(data, 'w')

        for d in info:
            path = d.get('path', d.get('id'))
            filename = os.path.basename(path)
            dir_path = os.path.dirname(path)

            # Write data
            _d = d.get('data', '')
            fpath = os.path.join(dir_path, filename)
            out.writestr(fpath, _d)

            metadata = d.copy()
            for name in ('data', 'path'):
                if name in metadata:
                    del metadata[name]
            # Write metadata
            metadata_path = os.path.join(dir_path, '.metadata')
            fpath = os.path.join(metadata_path, filename)
            _d = str(self.atxml_template(info=metadata))
            out.writestr(fpath, _d)

        out.close()
        data.seek(0)
        return data

InitializeClass(Export)



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
"""

from zope.interface import implementer

from Products.GenericSetup.interfaces import IFilesystemExporter
from Products.GenericSetup.interfaces import IFilesystemImporter
from Products.GenericSetup.utils import ExportConfiguratorBase
from Products.GenericSetup.utils import ImportConfiguratorBase
from Products.GenericSetup.utils import DEFAULT
from Products.GenericSetup.utils import KEY
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Products.Marshall.config import TOOL_ID
from Products.Marshall.registry import createPredicate

_FILENAME = 'marshall-registry.xml'


def _getRegistry(site):
    registry = site._getOb(TOOL_ID, None)

    if registry is None:
        raise ValueError('Marshall Registry Not Found')

    return registry


def exportMarshallRegistry(context):
    """ Export marshall registry as an XML file.

    o Designed for use as a GenericSetup export step.
    """
    registry = _getRegistry(context.getSite())
    exporter = MarshallRegistryExporter(registry).__of__(registry)
    xml = exporter.generateXML()
    context.writeDataFile(_FILENAME, xml, 'text/xml')
    return 'Marshall Registry exported.'


def _updateMarshallRegistry(registry, xml, should_purge, encoding=None):

    if should_purge:

        registry.manage_delObjects(ids=list(registry.objectIds()))

    importer = MarshallRegistryImporter(registry, encoding)
    reg_info = importer.parseXML(xml)

    for info in reg_info['predicates']:
        if registry.hasObject(info['id']):
            registry.manage_delObjects(ids=[info['id']])
        obj = createPredicate(info['predicate_name'],
                              info['id'], info['title'],
                              info['expression'],
                              info['component'])
        if (info['predicate_name'] in ('xmlns_attr',) and
            info.get('xml_settings', ())):
            obj.edit(**info['xml_settings'][0])
        registry._setObject(info['id'], obj)


def importMarshallRegistry(context):
    """ Import marshall registry from an XML file.

    o Designed for use as a GenericSetup import step.
    """
    registry = _getRegistry(context.getSite())
    encoding = context.getEncoding()

    xml = context.readDataFile(_FILENAME)
    if xml is None:
        return 'Site properties: Nothing to import.'

    _updateMarshallRegistry(registry, xml, context.shouldPurge(), encoding)
    return 'Marshall Registry imported.'


class MarshallRegistryExporter(ExportConfiguratorBase):

    def __init__(self, context, encoding='utf-8'):
        ExportConfiguratorBase.__init__(self, None, encoding)
        self.context = context

    def _getExportTemplate(self):
        return PageTemplateFile('xml/mrExport.xml', globals())

    def listPredicates(self):
        for predicate in self.context.objectValues():
            info = {}
            info['id'] = predicate.getId()
            info['title'] = predicate.title_or_id()
            info['expression'] = predicate.getExpression()
            info['component'] = predicate.getComponentName()
            info['predicate_name'] = predicate.getPredicateName()
            if info['predicate_name'] in ('xmlns_attr',):
                info['xml_settings'] = settings = {}
                settings['element_name'] = predicate.getElementName()
                settings['element_ns'] = predicate.getElementNS()
                settings['attr_name'] = predicate.getAttributeName()
                settings['attr_ns'] = predicate.getAttributeNS()
                settings['value'] = predicate.getValue()
            yield info


class MarshallRegistryImporter(ImportConfiguratorBase):

    def __init__(self, context, encoding=None):
        ImportConfiguratorBase.__init__(self, None, encoding)
        self.context = context

    def _getImportMapping(self):

        return {
          'marshall-registry':
              {'predicate':        {KEY: 'predicates', DEFAULT: ()},
               },
          'predicate':
              {'id':               {KEY: 'id'},
               'title':            {KEY: 'title'},
               'expression':       {KEY: 'expression'},
               'component':        {KEY: 'component'},
               'predicate_name':   {KEY: 'predicate_name'},
               'xml-settings':     {KEY: 'xml_settings', DEFAULT: ()}
               },
          'xml-settings':
              {'element-name':   {KEY: 'element_name'},
               'element-ns':     {KEY: 'element_ns'},
               'attr-name':      {KEY: 'attr_name'},
               'attr-ns':        {KEY: 'attr_ns'},
               'value':          {KEY: 'value'}
               },
         }


@implementer(IFilesystemExporter, IFilesystemImporter)
class MarshallRegistryFileExportImportAdapter(object):
    """ Designed for use when exporting / importing within a container.
    """

    def __init__(self, context):
        self.context = context

    def export(self, export_context, subdir, root=False):
        """ See IFilesystemExporter.
        """
        context = self.context
        exporter = MarshallRegistryExporter(context).__of__(context)
        xml = exporter.generateXML()
        export_context.writeDataFile(_FILENAME,
                                     xml,
                                     'text/xml',
                                     subdir,
                                    )

    def listExportableItems(self):
        """ See IFilesystemExporter.
        """
        return ()

    def import_(self, import_context, subdir, root=False):
        """ See IFilesystemImporter.
        """
        data = import_context.readDataFile(_FILENAME, subdir)
        if data is None:
            import_context.note('SGAIFA',
                                'no %s in %s' % (_FILENAME, subdir))
        else:
            pass



# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004-2006 Enfold Systems, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
$Id$
"""

from xml.dom import minidom
from AccessControl import ClassSecurityInfo
from App.class_init import InitializeClass
from Products.CMFCore.permissions import ManagePortal
from _base import Predicate
from Products.Marshall.registry import registerPredicate
from Products.PageTemplates.PageTemplateFile import PageTemplateFile


def attrMethod(attr, default=''):
    def boundAttrMethod(self):
        return getattr(self, attr, default)
    return boundAttrMethod


class XMLNS(Predicate):
    """ Predicate matching xml files on a xmlns:attribute pair.
    """

    security = ClassSecurityInfo()
    security.declareProtected(ManagePortal, 'edit',
                              'getElementNS', 'getElementName',
                              'getAttributeNS', 'getAttributeName',
                              'getValue')
    getElementNS = attrMethod('_element_ns')
    getElementName = attrMethod('_element_name')
    getAttributeNS = attrMethod('_attr_ns')
    getAttributeName = attrMethod('_attr_name')
    getValue = attrMethod('_value', None)

    manage_options = (
        Predicate.manage_options[0],
        {'label': 'Settings', 'action': 'manage_changeSettingsForm'},
        ) + Predicate.manage_options[1:]

    security.declareProtected('View management screens',
                              'manage_changeSettingsForm')

    manage_changeSettingsForm = PageTemplateFile(
        '../www/xmlnsSettings', globals(),
        __name__='manage_changeSettingsForm')

    manage_changeSettingsForm._owner = None

    security.declareProtected(ManagePortal, 'edit')
    def edit(self, element_ns='', element_name='',
             attr_ns='', attr_name='', value='', REQUEST=None):
        """Change settings germane to this predicate
        """

        for n in (element_ns, element_name, attr_ns, attr_name):
            if not isinstance(n, basestring):
                raise TypeError('string required, got %r.' % n)
        self._value = value
        self._element_ns = element_ns
        self._element_name = element_name
        self._attr_ns = attr_ns
        self._attr_name = attr_name

        if REQUEST is not None:
            message = 'Predicate Settings Changed.'
            return self.manage_changeSettingsForm(
                manage_tabs_message=message,
                management_view='Settings')

    security.declareProtected(ManagePortal, 'apply')
    def apply(self, obj, **kw):
        """ Return component name if the rule matches, else an empty tuple.
        """
        retval = Predicate.apply(self, obj, **kw)
        if not retval:
            return ()
        body = kw.get('data')
        if not body:
            f = kw.get('file')
            if f is None:
                return ()
            # Try 'file', though that may bloat memory with big files :(
            body = f.read()
            f.seek(0)
            if not body:
                return ()
        try:
            doc = minidom.parseString(body)
        except:
            return ()
        elm_args = filter(None, (self.getElementNS(), self.getElementName()))
        get_elm = (len(elm_args) == 2 and doc.getElementsByTagNameNS
                   or doc.getElementsByTagName)
        elm = get_elm(*elm_args)
        if not elm:
            return ()
        match = True
        elm = elm[0]
        attr_args = filter(None, (self.getAttributeNS(),
                                  self.getAttributeName()))
        if not attr_args:
            get_attr = lambda elm=elm: elm.firstChild.nodeValue.strip()
        else:
            get_attr = (len(attr_args) == 2 and elm.getAttributeNS or
                        elm.getAttribute)
        expected = self.getValue()
        if expected is not None:
            got = get_attr(*attr_args)
            match = got == expected
        return match and retval or ()

InitializeClass(XMLNS)
registerPredicate('xmlns_attr', 'XMLNS Element/Attribute', XMLNS)



##################################################################
# Marshall: A framework for pluggable marshalling policies
# Copyright (C) 2004 EnfoldSystems, LLC
# Copyright (C) 2004 ObjectRealms, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

"""
"""
from Products.Marshall.handlers.atxml import XmlNamespace
from Products.Marshall.handlers.atxml import SchemaAttribute

RNGSchemaFragment = '''
  <define name="DublinCore"
          xmlns:dc="http://purl.org/dc/elements/1.1/"
          datatypeLibrary="http://www.w3.org/2001/XMLSchema-datatypes"
          xmlns="http://relaxng.org/ns/structure/1.0">
    <choice>
      <zeroOrMore>
        <element name="dc:title"><text /></element>
      </zeroOrMore>
      <zeroOrMore>
        <element name="dc:description"><text /></element>
      </zeroOrMore>
      <zeroOrMore>
        <element name="dc:subject"><text /></element>
      </zeroOrMore>
      <zeroOrMore>
        <element name="dc:contributor"><data type="NMTOKEN" /></element>
      </zeroOrMore>
      <oneOrMore>
        <element name="dc:creator"><data type="NMTOKEN" /></element>
      </oneOrMore>
      <zeroOrMore>
        <element name="dc:language"><data type="language" /></element>
      </zeroOrMore>
      <zeroOrMore>
        <element name="dc:rights"><text /></element>
      </zeroOrMore>
    </choice>
  </define>
'''


class DCAttribute(SchemaAttribute):

    def __init__(self,
                 name,
                 accessor,
                 mutator,
                 many=False,
                 process=()):

        super(DCAttribute, self).__init__(name)

        self.accessor = accessor
        self.mutator = mutator
        self.many = many
        self.process = process

    def get(self, instance):
        accessor = getattr(instance, self.accessor)
        value = accessor()

        if not value:
            return None
        elif not isinstance(value, (list, tuple)):
            values = [value]
        elif self.many:
            values = value
        else:
            raise AssertionError("Many values on single value attr")

        return filter(None, values)

    def deserialize(self, instance, ns_data):

        value = ns_data.get(self.name)

        if not self.mutator or not value:
            return

        if self.process:
            for p in self.process:
                value = p(value)

        mutator = getattr(instance, self.mutator)
        mutator(value)

    def serialize(self, dom, parent_node, instance):
        values = self.get(instance)
        if not values:
            return False

        for value in values:
            elname = "%s:%s" % (self.namespace.prefix, self.name)
            node = dom.createElementNS(DublinCore.xmlns, elname)
            value_node = dom.createTextNode(str(value))
            node.appendChild(value_node)
            node.normalize()
            parent_node.appendChild(node)
        return True

    def processXmlValue(self, context, value):
        value = value and value.strip()
        if not value:
            return
        data = context.getDataFor(self.namespace.xmlns)

        if self.many:
            data.setdefault(self.name, []).append(value)
        else:
            data[self.name] = value


class normalizer(object):
    """ utility function ns """

    def space(text):
        return '\n'.join([s.strip() for s in text.splitlines()])
    space = staticmethod(space)

    def newline(text):
        return ' '.join([s.strip() for s in text.splitlines()])
    newline = staticmethod(newline)


class DublinCore(XmlNamespace):

    xmlns = 'http://purl.org/dc/elements/1.1/'
    prefix = 'dc'

    uses_at_fields = True

    attributes = (
        DCAttribute('title', 'Title', 'setTitle',
                    process=(normalizer.space, normalizer.newline)),

        DCAttribute('description', 'Description', 'setDescription',
                    process=(normalizer.space,)),

        DCAttribute('subject', 'Subject', 'setSubject', many=True),
        DCAttribute('contributor', 'Contributors', 'setContributors',
                    many=True),
        # this attr diverges from cmfdefault.dublincore
        DCAttribute('creator', 'Creators', 'setCreators', many=True),
        DCAttribute('rights', 'Rights', 'setRights'),
        DCAttribute('language', 'Language', 'setLanguage'))

    def getATFields(self):
        return ('title',
                'description',
                'contributors',
                'subject',
                'creators',
                'rights',
                'language')

    def getSchemaInfo(self):
        return [("DublinCore", "zeroOrMore", RNGSchemaFragment)]
